#!/usr/bin/perl -T

# This is a library of functions used by Captrap's view generation programs.

# Copyright 2009 Corey Hickey


# This file is part of Captrap.
#
# Captrap is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Captrap is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Captrap.  If not, see <http://www.gnu.org/licenses/>.


use strict;
use warnings FATAL => 'all';
use Switch;

package Captrap::View;
require Exporter;
our @ISA = ('Exporter');
our @EXPORT = ();
our %EXPORT_TAGS = (
  view => [ qw(
    mk_view_info
    mk_views
    cgi_list_views
  ) ],
  param => [ qw(
    mk_param_info
  ) ],
);
# all the tagged functions
our @EXPORT_OK = map({ @$_ } values(%EXPORT_TAGS));

# for development using a different Captrap module
use lib "lib";
use Captrap qw(:misc :db);


# -----------------------------------------------------------------------------
# misc.
# -----------------------------------------------------------------------------

# initialize priv hash according to what keys need to be filled;
# keys with defined values are ignored
sub init_priv {
  my $common = shift;
  my $needed = shift;
  $common->{priv} = {} unless defined($common->{priv});
  my $priv = $common->{priv};
  foreach my $key (@{$needed}) {
    next if defined($priv->{$key}); # already done
    my $val;
    switch ($key) {
      case ("avg_bytes") {
        $val = calc_avg_bytes($common);
      }
      case ("times") {
        # $common->{priv}->{now} may be undef
        $val = get_times($common, $common->{priv}->{now});
      }
      case ("lengths") {
        init_priv($common, [ "times" ]);
        $val = get_lengths($common, $priv->{"times"}->{now});
      }
      case ("mtpred_strings") {
        init_priv($common, [ qw(avg_bytes intervals states) ]);
        $val = format_mtpred_strings($priv->{avg_bytes}, $priv->{intervals},
            $priv->{states});
      }
      case ("states") {
        $val = [ qw(total up down bcast bogey) ];
      }
      case ("intervals") {
        $val = [ qw(year month day hour) ];
      }
      case ("bytes_current") {
	init_priv($common, [ qw(times states) ]);
        $val = get_state_bytes_from_current_hour($common,
            $priv->{times}->{now});
      }
      case ("bytes_last") {
        init_priv($common, [ qw(times) ]);
        $val = get_bytes_from_last_intervals($common,
            floor_unit("hour", $priv->{"times"}->{now}));
      }
      case ("per_starts") {
        $val = {
          year  => "first",
          month => "lastyear",
          day   => "lastmonth",
          hour  => "lastday",
        };
      }
      case ("unknown_macs") {
        $val = get_unknown_macs($common);
      }
      else { die "init_priv: don't know what to do with key '$key'" };
    }
    $priv->{$key} = $val;
  }
  return $priv;
}


# round to nearest int
sub round {
  my $number = shift;
  return int($number + 0.5);
}


# calculate the average bytes per day for each state and interval
sub calc_avg_bytes {
  my $common = shift;
  my $priv = init_priv($common, [ qw(bytes_last lengths states intervals) ]);
  my $bytes = $priv->{bytes_last};
  my $lengths = $priv->{lengths};
  my $states = $priv->{states};
  my $intervals = $priv->{intervals};
  my $avg = {};
  foreach my $interval (@{$intervals}) {
    foreach my $state (@{$states}) {
      $avg->{$interval}->{$state} = round(
          $bytes->{$interval}->{$state} / $lengths->{$interval});
    }
  }
  return $avg;
}

# -----------------------------------------------------------------------------
# formatting of data for CGI
# -----------------------------------------------------------------------------

# make a table row out of bytes/state data
sub mk_bytes_table_row {
  my $cgi = shift;
  my $bytes = shift; # hash ref
  my $states = shift; # array ref
  my $row_label = shift;
  my $row = $cgi->td($row_label);
  foreach my $state (@{$states}) {
    $row .= $cgi->td(format_bytes($cgi, $bytes->{$state}));
  }
  return $cgi->Tr($row);
}


# make a table out of bytes/interval/state data
sub mk_bytes_table {
  my $cgi = shift;
  my $bytes = shift; # hash ref
  my $intervals = shift; # array ref
  my $states = shift; # array ref
  # first make the first row of labels
  my $table = $cgi->td("");
  foreach my $state (@{$states}) {
    $table .= $cgi->td($state);
  }
  $table = $cgi->Tr($table);
  # now fill in the data rows
  foreach my $interval (@{$intervals}) {
    $table .= mk_bytes_table_row($cgi, $bytes->{$interval}, $states, $interval);
  }
  return $cgi->table({-border => 1}, $table);
}


# make strings of parameters for the mtpred graph
sub format_mtpred_strings {
  my $bytes = shift; # hash ref
  my $intervals = shift; # array ref
  my $states = shift; # array ref
  my $mtpred = {};
  foreach my $state (@{$states}) {
    my $string;
    foreach my $interval (@{$intervals}) {
      # make something like "12345:last_day,"
      $string .= $bytes->{$interval}->{$state} . ":last_" . $interval . ',';
    }
    # cut off the last ','
    $mtpred->{$state} = substr($string, 0, -1);
  }
  return $mtpred
}


# make an image with a border
sub mk_graph_image {
  my $common = shift; # hash ref
  my $src = shift;
  my ($w, $h) = @{$common->{config}}{"default_graph_w", "default_graph_h"};
  my $cgi = $common->{cgi};
  return $cgi->p($cgi->img({'src' => $src, 'border' => 1, 'width' => $w,
      'height' => $h}));
}


# wrapper for making the usual graphs per interval
sub mk_graph_links_per {
  my $common = shift; # hash ref
  my $states = shift; # may be undef
  my $cgi = $common->{cgi};
  my $priv = init_priv($common, [ qw(intervals per_starts) ]);
  my $text;
  foreach my $interval (@{$priv->{intervals}}) {
    my $start = $priv->{per_starts}->{$interval};
    my $src = mk_graph_link($common, "per", $interval, $start, $states);
    $text .= mk_graph_image($common, $src);
  }
  return $text;
}


# wrapper for making the usual moving total graphs
sub mk_graph_links_mt {
  my $common = shift; # hash ref
  my $states = shift; # may be undef
  my $cgi = $common->{cgi};
  my $priv = init_priv($common, [ qw(intervals per_starts) ]);
  my $text;
  # skip the first interval (year)
  for (my $i = 1; $i < @{$priv->{intervals}}; ++$i) {
    my $interval = $priv->{intervals}->[$i];
    my $start = $priv->{per_starts}->{$interval};
    my $sum_unit = $priv->{intervals}->[$i-1];
    my $extra = { sum_unit => [ $sum_unit ], sum_interval => [ 1 ] };
    my $src = mk_graph_link($common, "mt", $interval, $start, $states, $extra);
    $text .= mk_graph_image($common, $src);
  }
  return $text;
}


# make some mtpred graphs
sub mk_graph_links_mtpred {
  my $common = shift;
  my $states = shift; # array ref; may be undef
  my $priv = init_priv($common, [ "mtpred_strings" ]);
  unless (@$states) {
    $states = [ "total" ];
  }
  my $text;
  foreach my $state (@$states) {
    my $mtpred_bytes = $priv->{mtpred_strings}->{$state};
    $state = undef if $state eq "total";
    my $link = mk_graph_link($common, "mtpred", "day", "lastmonth", $state,
        { bytes => [ $mtpred_bytes] });
    $text .= mk_graph_image($common, $link);
  }
  return $text;
}


# make a link to a graph image
# end time is hardcoded to "now"; if I need a more general function later I'll
# write it then.
sub mk_graph_link {
  my $common = shift;
  my $graph = shift;
  my $per = shift;
  my $start = shift;
  my $states = shift;
  my $extra = shift;
  my $priv = init_priv($common, [ qw(times) ]);
  $extra = {} unless defined($extra);
  my $params = mk_ixhash();
  %$params = (
    graph => [ $graph],
    per => [ $per ],
    start => [ $priv->{"times"}->{$start} ],
    end => [$priv->{"times"}->{now} ],
    %$extra,
  );
  if (defined($states)) {
    $params->{states} = [ $states ];
  }
  my $caps = get_caps($common, $states, $params);
  if (defined($caps)) {
    $params->{caps} = [ $caps ];
  }
  my $grapher = $common->{config}->{grapher};
  return mk_link($common, $grapher, $params);
}


# get a list of transfer caps according to states and $per
sub get_caps {
  my $common = shift;
  my $states = shift; # array ref
  my $params = shift; # hash ref
  my $caps = $common->{config}->{caps};
  my $cap_unit;
  switch ($params->{graph}) {
    case (/^per$/i) {
      $cap_unit = $params->{per};
    }
    case (/^mt$/i) {
      $cap_unit = $params->{sum_unit};
    }
    case (/^mtpred$/i) {
      $cap_unit = "month";
    }
  }
  my $pos;
  switch ($cap_unit) {
    case (/^year$/i) {
      $pos = 0;
    }
    case (/^month$/i) {
      $pos = 1;
    }
    case (/^day$/i) {
      $pos = 2;
    }
    case (/^hour$/i) {
      $pos = 3;
    }
    else {
      die "unrecognized cap unit '$cap_unit'";
    }
  }
  unless (defined($states)) {
    return $caps->{total}->[$pos];
  }
  my @list;
  my $empty = 1;
  foreach my $state (l2a($states)) {
    # year,month,day,hour
    my $cap = $caps->{$state}->[$pos];
    if (defined($cap)) {
      push(@list, $cap);
      $empty = 0;
    } else {
      # use a -1 to take up space in case we get a real value later
      push(@list, -1);
    }
  }
  return undef if $empty;
  return join(',', @list);
}


# return a string to print
sub format_bytes {
  my $cgi = shift;
  my $bytes = shift;
  my ($factor, $suffix) = choose_factor($bytes);
  my $adjusted = $bytes / $factor;
  # 32-bit perl's sprintf doesn't seem to understand "%lld" properly,
  # so just let $bytes go in the format string
  return sprintf("$bytes%s(%.02f %s)", $cgi->br, $adjusted, $suffix);
}


# -----------------------------------------------------------------------------
# database fetching
# -----------------------------------------------------------------------------

# get lengths (in days) of various intervals
sub get_lengths {
  my $common = shift;
  my $now = shift;
  # days/year and days/month are variable,
  # days/day and days/hour should be constant
  my $sel = "
    SELECT
      DATEDIFF(?, ? - INTERVAL 1 YEAR),
      DATEDIFF(?, ? - INTERVAL 1 MONTH),
      1,
      1/24
  ";
  my $sth = $common->{dbh}->prepare($sel);
  $sth->execute($now, $now, $now, $now);
  my @a = $sth->fetchrow_array();
  die unless @a; # shouldn't happen
  return {
    year  => $a[0],
    month => $a[1],
    day   => $a[2],
    hour  => $a[3],
  };
}


# Get the bytes transferred over the most recent hour/day/month/year intervals.
# returns a hash reference:
# $bytes = {
#   interval1 => {
#     state1  => 1234,
#     state2  => 456,
#   },
#   interval2 => {
#     state1 => 7890,
#     state2 => 1234,
#   },
#   ...
# }
sub get_bytes_from_last_intervals {
  my $common = shift;
  my $end = shift;
  my $priv = init_priv($common, [ qw(states intervals) ]);
  my $bytes = {};
  foreach my $interval (@{$priv->{intervals}}) {
    $bytes->{$interval} =
        get_bytes_from_interval($common, $end, 1, $interval, $priv->{states});
  }
  return $bytes;
}


# get the bytes transfered within an interval, separated by state
# returns a hash reference
# $bytes = {
#   state1 => 1234,
#   state2 => 4567,
#   ...
#   total  => 141234,
# }
sub get_bytes_from_interval {
  my $common = shift;
  my $end = shift;
  my $num = shift;
  my $interval = shift;
  my $states = shift;
  my $bytes = {};
  foreach my $state (@$states) {
    # default 0 bytes
    $bytes->{$state} = 0;
  }
  die "invalid interval unit" unless check_interval_unit($interval);
  my ($acct, $macs) = @{$common->{config}}{"db_table_acct","db_table_macs"};
  my $states_ignore = $common->{config}->{states_ignore};
  my $ph = placeholder_list($states_ignore);
  my $sel = "
    SELECT $macs.state, sum($acct.bytes)
    FROM $acct JOIN $macs
      ON ($acct.mac_dst = $macs.mac)
    WHERE
      $acct.stamp_inserted < ? AND
      $acct.stamp_inserted >= ? - INTERVAL ? $interval AND
      $macs.state NOT IN ($ph)
    GROUP BY $macs.state
  ";
  my $sth = $common->{dbh}->prepare($sel);
  $sth->execute($end, $end, $num, @$states_ignore);
  $bytes->{total} = 0;
  while (my @a = $sth->fetchrow_array()) {
    my ($state, $sum) = (@a);
    # don't record states that weren't specified, but do add them to the total
    $bytes->{total} += $sum;
    next unless defined($bytes->{$state});
    $bytes->{$state} = $sum;
  }
  return $bytes;
}


# fetch a single value; return 0 if the value returned would be undef
sub fetch_val_or_zero {
  my $sth = shift; # should already be executed
  my @a = $sth->fetchrow_array();
  die unless @a; # shouldn't happen
  # a null value means 0 bytes
  return defined($a[0]) ? $a[0] : 0;
}


# get the bytes transferred during the current hour, for each state
sub get_state_bytes_from_current_hour {
  my $common = shift;
  my $priv = $common->{priv};
  my $now = shift;
  my $states = $priv->{states};
  my $start = floor_unit("hour", $now);
  my $bytes = {};
  foreach my $state (@{$states}) {
    $bytes->{"current"}->{$state} =
        get_bytes_between($common, $start, $now, $state);
  }
  return $bytes;
}


# get the total bytes transfered between two times
sub get_bytes_between {
  my $common = shift;
  my $start = shift;
  my $end = shift;
  my $state = shift; # may be undef
  my $partial = ! safe_eq($state, "total");
  my ($acct, $macs) = @{$common->{config}}{"db_table_acct","db_table_macs"};
  my $sel = "
    SELECT SUM(bytes)
    FROM $acct
    WHERE stamp_inserted >= ? AND stamp_inserted < ?
  ";
  if ($partial) {
    $sel = "
      SELECT SUM($acct.bytes)
      FROM $acct JOIN $macs
        ON ($acct.mac_dst = $macs.mac)
      WHERE stamp_inserted >= ? AND stamp_inserted < ? AND $macs.state = ?";
  }
  my $sth = $common->{dbh}->prepare($sel);
  if ($partial) {
    $sth->execute($start, $end, $state);
  } else {
    $sth->execute($start, $end);
  }
  return fetch_val_or_zero($sth);
}

# -----------------------------------------------------------------------------
# view generation
# -----------------------------------------------------------------------------

# make a dispatch table relating views to functions
sub mk_view_funcs {
  return {
    basic              => \&mk_view_basic,
    graph_inter        => \&mk_view_graph_inter,
    graph_inter_updown => \&mk_view_graph_inter_updown,
    graph_inter_bb     => \&mk_view_graph_inter_bb,
    graph_adv          => \&mk_view_graph_adv,
    graph_adv_updown   => \&mk_view_graph_adv_updown,
    graph_adv_bb       => \&mk_view_graph_adv_bb,
  }
}


# make a hash of available views
sub mk_view_info {
  my $views = mk_ixhash();
  %$views = (
    basic              => {
      ttl => "Basic Information",
      txt => "tables of transfer information",
    },
    graph_inter        => {
      ttl => "Interval Graphs",
      txt => "graphs of traffic transferred over last intervals",
    },
    graph_inter_updown => {
      ttl => "Interval Graphs (up/down)",
      txt => "interval graphs for upload and download traffic",
    },
    graph_inter_bb     => {
      ttl => "Interval Graphs (bcast/bogey)",
      txt => "interval graphs for broadcast and bogey traffic",
    },
    graph_adv          => {
      ttl => "Advanced Graphs",
      txt => "advanced graphs",
    },
    graph_adv_updown   => {
      ttl => "Advanced Graphs (up/down)",
      txt => "advanced graphs for upload and download traffic",
    },
    graph_adv_bb       => {
      ttl => "Advanced Graphs (bcast/bogey)",
      txt => "advanced graphs for broadcast and bogey traffic",
    },
  );
  return $views;
}


# make one or more views
sub mk_views {
  my $common = shift; # hash ref
  my $params = shift; # hash ref
  my $extension = shift; # scalar ref
  my $cgi = $common->{cgi};
  $$extension = "html";
  # must set this before init_priv() gets a chance to
  if (defined($params->{now})) {
    $common->{priv}->{now} = $params->{now};
  }
  my $views = mk_view_info();
  my $funcs = mk_view_funcs();
  my $text;
  $text .= $cgi->start_html('-title' => 'Captrap');
  $text .= $cgi->h1("Selected Captrap Views");
  foreach my $view (l2a($params->{views})) {
    if ($view eq "all") {
      foreach my $v (keys %$views) {
        $text .= $cgi->h2($views->{$v}->{ttl});
        $text .= &{$funcs->{$v}}($common);
      }
    } elsif ($view eq "list") {
      $text .= cgi_list_views($common, $views);
    } elsif (defined($views->{$view})) {
      $text .= $cgi->h2($views->{$view}->{ttl});
      $text .= &{$funcs->{$view}}($common);
    } else {
      mk_view_error($common->{cgi}, $view);
    }
  }
  $text .= $cgi->end_html();
  return $text;
}


# list all views in cgi
sub cgi_list_views {
  my $common = shift;
  my $views = shift; # hash ref
  my $cgi = $common->{cgi};
  my $text;
  $text .= $cgi->h2('Supported Views');
  $text .= $cgi->p("Multiple views can be retrieved in the same request by
      joining them with a comma. For example: \"views=basic,adv\"");
  $text .= $cgi->h3('Regular Views');
  foreach my $view (keys %$views) {
    $text .= $cgi->h5($view);
    $text .= $cgi->p($views->{$view}->{txt});
  }
  $text .= $cgi->h3('Meta-views');
  $text .= $cgi->h5("all");
  $text .= $cgi->p("all supported views (slow, useful for debug)");
  $text .= $cgi->h5("list");
  $text .= $cgi->p("this list of views");
  return $text;
}


# make the basic information view
sub mk_view_basic {
  my $common = shift;
  my $cgi = $common->{cgi};
  my $needed = [ qw(
    avg_bytes
    bytes_current
    bytes_last
    intervals
    states
    times
    unknown_macs
   ) ];
  my $priv = init_priv($common, $needed);
  my $text;
    
  if (defined($priv->{now})) {
    $text .= $cgi->p("Specified current time: "
        . print_time($priv->{"times"}->{now}));
  } else {
    $text .= $cgi->p("Current time from database server: "
        . print_time($priv->{"times"}->{now}));
  }
  $text .= $cgi->p("Statistics are listed for interface " .
      "$common->{config}->{interface}.");

  $text .= $cgi->h4('Unknown MAC Addresses');
  my $macs = $priv->{unknown_macs};
  if (@$macs) {
    # todo: how?
    $text .= $cgi->p("The following MAC addresses are unrecognized. They " .
      "will be accounted for in totals but not in individual traffic states.");
    my $table = $cgi->Tr($cgi->td("MAC"), $cgi->td("bytes"),
        $cgi->td("first seen"), $cgi->td("last seen"));
    foreach my $row (@$macs) {
      my $tr;
      foreach my $td (@$row) {
        $tr .= $cgi->td($td);
      }
      $table .= $cgi->Tr($tr);
    }
    $text .= $cgi->table({-border => 1}, $table);
  } else {
    $text .= $cgi->p("No unknown MAC addresses found (this is good).");
  }

  # if "now" is specified, it gets rounded down, so there won't be any
  # current bytes
  unless (defined($priv->{now})) {
    $text .= $cgi->h4("Bytes Transferred During Current Hour");
    $text .= mk_bytes_table($cgi, $priv->{bytes_current}, [ "current" ],
      $priv->{states});
  }
  $text .= $cgi->h4("Bytes Transferred Over Last Intervals");
  unless (defined($priv->{now})) {
    $text .= $cgi->p("Note: this does not include bytes transferred during " .
        "the current hour");
  }
  $text .= mk_bytes_table($cgi, $priv->{bytes_last}, $priv->{intervals},
      $priv->{states});
  $text .= $cgi->h4("Average Bytes per Day Over Last Intervals");
  $text .= mk_bytes_table($cgi, $priv->{avg_bytes}, $priv->{intervals},
      $priv->{states});
  return $text;
}


# make interval graphs
sub mk_view_graph_inter {
  my $common = shift;
  my $cgi = $common->{cgi};
  return mk_graph_links_per($common);
}


# make interval graphs with separate up/down bars
sub mk_view_graph_inter_updown {
  my $common = shift;
  my $cgi = $common->{cgi};
  return mk_graph_links_per($common, "up,down");
}


# make interval graphs with separate bcast/bogey bars
sub mk_view_graph_inter_bb {
  my $common = shift;
  my $cgi = $common->{cgi};
  return mk_graph_links_per($common, "bcast,bogey");
}


# make advanced graphs
sub mk_view_graph_adv {
  my $common = shift;
  return mk_graph_links_mt($common) .
      mk_graph_links_mtpred($common, [ ]);
}


# make advanced graphs with separate up/down bars
sub mk_view_graph_adv_updown {
  my $common = shift;
  return mk_graph_links_mt($common, "up,down") .
      mk_graph_links_mtpred($common, [ qw(up down) ]);
}


# make advanced graphs with separate bcast/bogey bars
sub mk_view_graph_adv_bb {
  my $common = shift;
  return mk_graph_links_mt($common, "bcast,bogey") .
      mk_graph_links_mtpred($common, [ qw(bcast bogey) ]);
}


# print a useful error when a view doesn't exist
sub mk_view_error {
  my $cgi = shift;
  my $view = shift;
  print $cgi->h2('Error');
  print $cgi->p("No such view: $view");
}

# -----------------------------------------------------------------------------
# param info
# -----------------------------------------------------------------------------

# set up defaults and regexes for parsing each parameter
# each parameter MUST have a key here
sub mk_param_info {
  my $info = mk_ixhash();
  # don't forget to start regexes with a ^ and end with a $
  %$info = (
    views => {
      def => "basic",
      # Just check for valid characters;
      # mk_views will handle unknown views gracefully
      reg => qr/^[[:word:],]{1,100}$/,
      txt => "Comma-separated list of views to display. Use 'views=list' " .
          "for information about available views.",
      var => 's',
    },
    now => {
      def => undef,
      reg => time_re("second", "T"),
      txt => "Current time. This will be rounded down to the nearest hour " .
          "and used as a maximum time for data retrieval.",
      var => 's',
    },
  );
  return $info;
}


1;
